/* UniThorpe <strick@yak.net> GnuGPL */

#include "unithorpe3.h"

typedef char* bc_proc(th G, int b);

bc_proc* bc_prox[ MAXINT+1 ];

static char* bc_wtf(th G, int bc)
{
    printf("WTF(\n");
	adump(G,1); 
    printf("WTF)\n");
	return NULL;
}

static char* bc_nop(th G, int bc)
{
	P("      NOP\n");
	return NULL;
}

//#define NEXT() ( printf("...next: '%c'\n",   aget( GetKP, aget( G, C_PC ) )), aget( GetKP, apostincr( G, C_PC ) ) )
//#define KODE(i) ( printf("........kode%d: '%c'\n", GetKP, aget( GetKP, i ) ), aget( GetKP, i ) )

#define NEXT() ( aget( GetKP, apostincr( G, C_PC ) ) )
#define KODE(i) ( aget( GetKP, i ) )

#define gget(X) aget( G, X )
#define gput(X,Y) aput( G, X, Y )

void push_my_var( th G, th var, th value )
{
	th sp= GetSP;
	apush( sp, gget( var ) );   // push old value
	apush( sp, var );		// push var name
	apush( sp, 101*'L' );  		// push L for Local
	gput( var, value );  	// set my var
}

static int match_char(int c)
{
	switch (c) {
		case '{': return '}';
		case '}': return '{';
		case '[': return ']';
		case ']': return '[';
		case '(': return ')';
		case ')': return '(';
		case '<': return '>';
		case '>': return '<';
		case '`': return '\'';
		case '\'': return '`';
	}
	return 0;
}

static char* bc_elihw(th G, int bc)
{
	th next;
	int end= KODE(GetPC-1); // look back at what char invoked us
	int begin= match_char(end);
	ASSERT( begin );

	next= NEXT();
	if (next<=32) {
		// we are at end -- skip back to begin
		th pc;
		for ( pc= GetPC-2; pc>0; pc-- ) {
			if ( KODE(pc) == '@' && KODE(pc+1) == begin ) {
				PutPC(pc+2);  // inside the first braces
				return NULL;
			}
		}
		return "bc_elihw-no-first-brace";
	} else {
		// check the var
		if ( gget( next) ) {
			// true -- continue
			next= NEXT();
			if (next!=begin) return "bc_elihw-bad-next-brace";
			return NULL;
		} else {
			// false -- break out
			th pc;
			next= NEXT();
			if (next!=begin) return "bc_elihw-bad-next-brace";

			for ( pc= GetPC+1; KODE(pc); pc++ ) {
				if ( KODE(pc) == end && KODE(pc+1) <= 32 ) {
					// we found the end
					PutPC(pc+2);  // beyond
					return NULL;
				}
			}
			return "bc_elihw-no-last-brace";
		}
	}
}
static char* bc_while(th G, int bc)
{
	th next= NEXT();
	int begin= INT( next );
	int end= match_char(begin);
	if (!end) return "bad-initial-match_char-in-while";

	push_my_var( G, end, C_ELIHW );
	return NULL;
}

static char* bc_fi(th G, int bc)
{
	th next;
	int end= KODE(GetPC-1); // look back at what char invoked us
	int begin= match_char(end);
	ASSERT( begin );

	next= NEXT();
	if (next==begin) {
			// skip over block
			th pc;
			for ( pc= GetPC+1; KODE(pc); pc++ ) {
				if ( KODE(pc) == end && KODE(pc+1) <= 32 ) {
					// we found the end
					PutPC(pc+2);  // beyond
					return NULL;
				}
			}
			return "bc_fi-no-last-brace";
	} else {
		// just keep going
		return NULL;
	}
}
static char* bc_if(th G, int bc)
{
	th next= NEXT();
	int var= INT( next );
	next= NEXT();
	int begin= INT( next );
	int end= match_char(begin);
	if (!end) return "bad-initial-match_char-in-if";

	push_my_var( G, end, C_FI );

	if (gget(var)) {
		// true -- keep going
		return NULL;
	} else {
		// false -- skip over block
		int pc;
		for ( pc= GetPC; KODE(pc); pc++ ) {
			if ( KODE(pc) == end ) {
				if ( KODE(pc+1) <= 32 ) {
					// we found the phrase end, there is no else block
					PutPC(pc+2);  // beyond
					return NULL;
				} else if ( KODE(pc+1) == begin ) {
					// jump into else block
					PutPC(pc+2);  // beyond
					return NULL;
				} else {
					return "bad-else-in-if";
				}
			}
		}
		return "no-end-to-if-then-block";
	}
}


static char* bc_lit_char(th G, int bc)
{
	int zv, av;
	zv= NEXT();
	if (zv<=32) return "missing-param-lit_char-z";
	av= NEXT();
	if (av==0) return "missing-param-lit_char-c";
	gput(zv,av);
	return NULL;
}


static char* bc_aput(th G, int bc)
{
	th xv, av, iv, a, x, i;
	av= NEXT();
	if (av<=32) return "missing-param-aput-a";
	iv= NEXT();
	if (iv<=32) return "missing-param-aput-i";
	xv= NEXT();
	if (xv<=32) return "missing-param-aput-x";

	a= gget( INT(av) );
	if (!ISPTR(a)) return "aput-a-not-array";
	i= gget( INT(iv) );
	if (!ISINT(i)) return "aput-i-not-int";
	x= gget( INT(xv) );

	aput(a,i,x); 
	return NULL;
}
static char* bc_aget(th G, int bc)
{
	th zv, av, iv, a, z, i;
	zv= NEXT();
	if (zv<=32) return "missing-param-aget-z";
	av= NEXT();
	if (av<=32) return "missing-param-aget-a";
	iv= NEXT();
	if (iv<=32) return "missing-param-aget-i";

	a= gget( INT(av) );
	if (!ISPTR(a)) return "aget-a-not-array";
	i= gget( INT(iv));
	if (!ISINT(i)) return "aget-i-not-int";

	z= aget(a,i); 
	gput(zv,z);
	return NULL;
}
static char* bc_anew(th G, int bc)
{
	th zv, z;
	zv= NEXT();
	if (zv<=32) return "missing-param-anew-z";

	gput(zv, anew() );
	return NULL;
}

static char* bc_relop(th G, int bc)
{
	int zv, av, bv, a, b, z;
	zv= NEXT();
	if (zv<=32) return "missing-param-relop-z";
	av= NEXT();
	if (av<=32) return "missing-param-relop-a";
	bv= NEXT();
	if (bv<=32) return "missing-param-relop-b";

	a= gget( INT(av) );
	b= gget( INT(bv) );
printf("RELOP--- %lu %c%c %lu ", a, hi(bc), lo(bc), b );
	switch (bc) {
	  case C_EQ: z= (a==b); break;
	  case C_NE: z= (a!=b); break;
	  case C_GT: z= (a> b); break;
	  case C_GE: z= (a>=b); break;
	  case C_LT: z= (a< b); break;
	  case C_LE: z= (a<=b); break;
	  default: return "bc_relop-default";
	}
	aput(G,zv,z);
printf("-> %lu\n", z );

	return NULL;
}

static char* bc_arith(th G, int bc)
{
	int zv, av, bv, a, b, z;
	zv= NEXT();
	if (zv<=32) return "missing-param-arith-z";
	av= NEXT();
	if (av<=32) return "missing-param-arith-a";
	bv= NEXT();
	if (bv<=32) return "missing-param-arith-b";

	a= INT( gget( INT(av)) );
	b= INT( gget( INT(bv)) );
	switch (bc) {
	  case '+': z= (a+b) & MAXINT; break;
	  case '-': z= (a-b) & MAXINT; break;
	  case '*': z= (a*b) & MAXINT; break;
	  case '/': z= (a/b) & MAXINT; break;
	  case '%': z= (a/b) & MAXINT; break;
	  case '&': z= (a&b) & MAXINT; break;
	  case '|': z= (a|b) & MAXINT; break;
	  case '^': z= (a^b) & MAXINT; break;
	  default: return "bc_arith-default";
	}
	aput(G,zv,z);

	return NULL;
}


void bc_init(th G) {

	setvbuf( stdout, NULL, _IONBF, 0 );

#define WTF HILO('\\','\\')
	bc_prox[ WTF ] = bc_wtf;	gput( WTF, WTF );


	gput( ';', ';' );   // magic return code

	bc_prox[ ' ' ] = bc_nop;	gput( ' ', ' ' ); // NOP


	bc_prox[ '+' ] = bc_arith;	gput( '+', '+' );
	bc_prox[ '-' ] = bc_arith;	gput( '-', '-' );
	bc_prox[ '*' ] = bc_arith;	gput( '*', '*' );
	bc_prox[ '/' ] = bc_arith;	gput( '/', '/' );
	bc_prox[ '%' ] = bc_arith;	gput( '%', '%' );
	bc_prox[ '&' ] = bc_arith;	gput( '&', '&' );
	bc_prox[ '|' ] = bc_arith;	gput( '|', '|' );
	bc_prox[ '^' ] = bc_arith;	gput( '^', '^' );

	bc_prox[ C_EQ ] = bc_relop;	gput( C_EQ, C_EQ );
	bc_prox[ C_NE ] = bc_relop;	gput( C_NE, C_NE );
	bc_prox[ C_GT ] = bc_relop;	gput( C_GT, C_GT );
	bc_prox[ C_GE ] = bc_relop;	gput( C_GE, C_GE );
	bc_prox[ C_LT ] = bc_relop;	gput( C_LT, C_LT );
	bc_prox[ C_LE ] = bc_relop;	gput( C_LE, C_LE );

	bc_prox[ '.' ] = bc_aput;	gput( '.', '.' );
	bc_prox[ ',' ] = bc_aget;	gput( ',', ',' );
	bc_prox[ '!' ] = bc_anew;	gput( '!', '!' );

	bc_prox[ '@' ] = bc_while;	gput( '@', '@' );
	bc_prox[ C_ELIHW ] = bc_elihw;	// not bound until bc_while

	bc_prox[ '?' ] = bc_if;		gput( '?', '?' );
	bc_prox[ C_FI ] = bc_fi;	// not bound until bc_if

        gput( '0', 0 );
        gput( '1', 1 );
        gput( '2', 2 );
        gput( '3', 3 );
        gput( '4', 4 );
        gput( '5', 5 );
        gput( '6', 6 );
        gput( '7', 7 );
        gput( '8', 8 );
        gput( '9', 9 );

	bc_prox[ ':' ] = bc_lit_char;     gput( ':', ':' );
	adump(G,0); 


}

int which_global(th G, th x)
{
	int i;
	for (i=33; i<126; i++) {
		if ( x == gget(i) ) {
			return i;
		}
	}
	return '?';
}


void bytecode(th G, int bc)
{
	bc_proc* pr = bc_prox[ INT(bc) ];

	P("BYTECODE '%c%c' [%c%2lu] bc=%d @%lu\n", hi(KODE(GetPC-1)), lo(KODE(GetPC-1)), which_global(G,GetKP), GetPC, bc, (th)pr );

	if (pr) { 
		char* e= pr(G, bc);
		if (e) bomb(e, bc);
	
		P("-----> ");
		adump(G,0);
		fflush(stdout);
	} else {
		bomb("null-bytecode-proc", bc);
	}
}

int loop(th G, int start) {
	int c,i;
	th sp;

P("*** BEGIN LOOP: G->"); adump(G,0);

	PutKP( gget( start)  );
	ASSERT( aget( GetKP, 0 ) == C_PROC );
	PutPC(1);

	sp= anew();
	PutSP( sp );
	apush( sp, 0 ); // begin stack with terminal 0

loop:
	while (1) {
		// advance PC and get code char c
		th cth = aget( GetKP, apostincr(G,C_PC) );
		c= INT(cth);
		
		if (!c) break; // if end off code line

		// lookup global object at c
		th b= gget(c);

		while ( ISPTR(b) ) {
			// we have an array.
			// if it starts with C_PROC, it is a proc.
			// if it starts with PTR, it is an object.
			if ( aget(b,0)==C_PROC || ISPTR(aget(b,0)) ) {

				th self= 0;
				th cls= 0;
				if ( ISPTR(aget(b,0)) ) {
				    self= b;
				    cls= aget(b,0);

				    cth= aget( GetKP, apostincr( G, C_PC ) );
				    c= INT( cth );
				    ASSERT(c);
				    b= aget( cls, c );

				    apush( sp, aget(G,'$') ); // push self
				    apush( sp, 101*'$' );     // know this is self
				    gput('$', self);
				    continue;
				}

				// remember where on stack this frame begins
				int frame= aget(sp,ASTACK_LEN_SLOT);
				int no_mo_args= 0;

				th oldPC= GetPC;
				th oldKP= GetKP;

				apush( sp, oldPC );
				apush( sp, oldKP );
				apush( sp, 101*C_PROC );
				
				// for each formal parameter
				for (i=1; aget(b,i); i++) {
					th arg;
					th p= aget(b,i);  // formal param p
					if (p<=32) break;
					INT(p);

					// get actual arg

					if ( !no_mo_args ) {
						arg= aget( oldKP, apostincr( sp, frame+ASTACK_BEGIN_SLOT ) );
						if (arg<=32) no_mo_args= 1;
						INT(arg);
					}

					if ( no_mo_args ) {
						// just save params as locals 
						apush( sp, gget( p ) );
						apush( sp, p );
						apush( sp, 101*'L' );  // L for Local
						gput( p, 0 );  // defaults to 0
					} else if (i==1) { // first param is in-out
						apush( sp, gget( p ) );
						apush( sp, p );
						// also push name of arg var
						apush( sp, arg );
						apush( sp, 101*'O' );  // O for in-Out
						gput( p, gget( arg) );  // set param to actual arg
					} else { // rest are in
						push_my_var( G, p, gget( arg) );
						//apush( sp, gget( p ) );
						////apush( sp, p );
						//apush( sp, 101*'L' );  // L for Local
						//gput( p, gget( arg) );  // set param to actual arg
					}
				}
				PutKP( b );
				PutPC( i );

P("*** CALLING [%c]%lu @%lu: G->", which_global(G,b), b, (ARR(sp))->p[ASTACK_LEN_SLOT] ); adump(G,0);
P("***         SP->" ); adump( sp,0 );
P("***         KP->" ); adump( GetKP,0 );
				goto loop;
			} else if (  aget(b,0)==0 ) {

			    //  it is a namespace array -- we traverse the slot
			    //	    // (namespace has 0 in slot 0)

			    c= INT( aget( GetKP, apostincr( G, C_PC ) ) );
			    ASSERT(c);
			    b= aget( b, c );

			} else {
			    bomb( "exec-arry-default", c );
			}
		}
		if (b==';') break;
if (b==0) {
	fprintf(stderr, "WARNING: b==0 c=%d kp=%lu pc=%lu\n", c, GetKP, GetPC );
}
		bytecode(G,b);
	}


	// end of script -- undo the stack
	while ((c = apop( sp ) )) {
		switch(c) {

		case 101*C_PROC: 
P("*** RETURNING @%lu: G->", sp ); adump(G,0);
			gput( C_KP, apop( sp ) );
			gput( C_PC, apop( sp ) );

			// tentative pop
			while (1) {
			    th maybe= apop(sp);
			    if ( maybe == 101*'$' ) {
				th old_self= apop(sp);
				gput( '$', old_self );
			    } else {
				// undo the pop by pushing
				apush( sp, maybe );
				break;
			    }
			}
			goto loop;
		case 101*'L': {
			th param= apop(sp); // name off param
			th saved= apop(sp); // saved value
			gput( param, saved );
			break;
		}

		case 101*'O': {
			th inout= apop(sp); // out param
			th param= apop(sp); // name of param
			th saved= apop(sp); // saved value
			if ( inout != param ) {
				gput( inout, gget( param)  );
				gput( param, saved );
			}
			break;
		}

		default:
			bomb( "pop-sp-default", c );

		}
	}
	P("*** END LOOP: G->");
	adump(G,0);
	return INT( gget( C_STAT ) );
}


		

		



